"""Command instruction messages (export)."""

from ...instruction import (
    ChannelMutingCommand as ChannelMuting,
    HDMIOutputCommand as HDMIOutput,
    InputSourceCommand as InputSource,
    ListeningModeCommand as ListeningMode,
    MutingCommand as Muting,
    PowerCommand as Power,
    ToneCommand as Tone,
    TunerPresetCommand as TunerPreset,
    VolumeCommand as Volume,
)



"""Messages."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, auto
from typing import TYPE_CHECKING, ClassVar, Self


class Zone(Enum):
    MAIN = "main"
    ZONE1 = MAIN
    ZONE2 = "zone2"
    ZONE3 = "zone3"
    ZONE4 = "zone4"

    def __repr__(self) -> str:
        return str(self)


ALL_ZONES = (Zone.MAIN, Zone.ZONE2, Zone.ZONE3, Zone.ZONE4)


class Kind(Enum):
    POWER = auto()
    MUTING = auto()
    CHANNEL_MUTING = auto()
    VOLUME = auto()
    TONE = auto()
    INPUT_SOURCE = auto()
    LISTENING_MODE = auto()
    HDMI_OUTPUT = auto()
    AUDIO_INFORMATION = auto()
    VIDEO_INFORMATION = auto()
    TUNER_PRESET = auto()
    FL_DISPLAY = auto()
    DISCOVERY = auto()

    def __repr__(self) -> str:
        return str(self)


@dataclass(slots=True, init=False)
class _CodeBaseConcreteMixin:
    kind: Kind
    zone: Zone
    raw: bytes = field(repr=False, init=False)

    def __new__(cls, kind: Kind, zone: Zone) -> Self:
        self = object.__new__(cls)
        self.kind = kind
        self.zone = zone
        return self

    @property
    def value(self) -> _CodeBaseConcreteMixin:
        """Code value."""
        # only for typing, this function is actually removed
        return self


del _CodeBaseConcreteMixin.value


# should be _CodeBase, but then the class vars don't work with Enum properly
class CodeBase(_CodeBaseConcreteMixin, Enum):
    __kind_mapping: ClassVar[dict[Kind, dict[Zone, Code]]] = {}
    __raw_mapping: ClassVar[dict[bytes, Code]] = {}

    def __init__(self, kind: Kind, zone: Zone) -> None:
        if TYPE_CHECKING:
            assert isinstance(self, Code)
        raw = self.name.encode()
        self.value.raw = self.raw = raw
        self.__kind_mapping.setdefault(kind, {})[zone] = self
        self.__raw_mapping[raw] = self

    @classmethod
    def from_kind_zone(cls, kind: Kind, zone: Zone) -> Code:
        return cls.__kind_mapping[kind][zone]

    @classmethod
    def get_from_kind_zone(cls, kind: Kind, zone: Zone) -> Code | None:
        return cls.__kind_mapping[kind].get(zone)

    @classmethod
    def parse(cls, raw_code: bytes) -> Code | None:
        return cls.__raw_mapping.get(raw_code)


class Code(CodeBase):
    # POWER
    PWR = Kind.POWER, Zone.MAIN
    ZPW = Kind.POWER, Zone.ZONE2
    PW3 = Kind.POWER, Zone.ZONE3
    PW4 = Kind.POWER, Zone.ZONE4
    # MUTING
    AMT = Kind.MUTING, Zone.MAIN
    ZMT = Kind.MUTING, Zone.ZONE2
    MT3 = Kind.MUTING, Zone.ZONE3
    MT4 = Kind.MUTING, Zone.ZONE4
    # CHANNEL MUTING
    CMT = Kind.CHANNEL_MUTING, Zone.MAIN
    # VOLUME
    MVL = Kind.VOLUME, Zone.MAIN
    ZVL = Kind.VOLUME, Zone.ZONE2
    VL3 = Kind.VOLUME, Zone.ZONE3
    VL4 = Kind.VOLUME, Zone.ZONE4
    # TONE
    TFR = Kind.TONE, Zone.MAIN
    ZTN = Kind.TONE, Zone.ZONE2
    TN3 = Kind.TONE, Zone.ZONE3
    # INPUT SOURCE
    SLI = Kind.INPUT_SOURCE, Zone.MAIN
    SLZ = Kind.INPUT_SOURCE, Zone.ZONE2
    SL3 = Kind.INPUT_SOURCE, Zone.ZONE3
    SL4 = Kind.INPUT_SOURCE, Zone.ZONE4
    # LISTENING MODE
    LMD = Kind.LISTENING_MODE, Zone.MAIN
    LMZ = Kind.LISTENING_MODE, Zone.ZONE2
    # HDMI OUTPUT
    HDO = Kind.HDMI_OUTPUT, Zone.MAIN
    # AUDIO INFORMATION
    IFA = Kind.AUDIO_INFORMATION, Zone.MAIN
    # VIDEO INFORMATION
    IFV = Kind.VIDEO_INFORMATION, Zone.MAIN
    # TUNER PRESET
    PRS = Kind.TUNER_PRESET, Zone.MAIN
    PRZ = Kind.TUNER_PRESET, Zone.ZONE2
    PR3 = Kind.TUNER_PRESET, Zone.ZONE3
    PR4 = Kind.TUNER_PRESET, Zone.ZONE4
    # FL DISPLAY
    FLD = Kind.FL_DISPLAY, Zone.MAIN
    # DISCOVERY
    ECN = Kind.DISCOVERY, Zone.MAIN


__all__ = [
    "Zone",
    "Kind",
    "Code",
]



"""Query instruction messages (export)."""

from ...instruction import (
    AudioInformationQuery as AudioInformation,
    ChannelMutingQuery as ChannelMuting,
    DiscoveryQuery as Discovery,
    FLDisplayQuery as FLDisplay,
    HDMIOutputQuery as HDMIOutput,
    InputSourceQuery as InputSource,
    ListeningModeQuery as ListeningMode,
    MutingQuery as Muting,
    PowerQuery as Power,
    ToneQuery as Tone,
    TunerPresetQuery as TunerPreset,
    VideoInformationQuery as VideoInformation,
    VolumeQuery as Volume,
)



"""Protocol."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from dataclasses import dataclass, field
from enum import Enum, auto
from functools import cached_property
import logging
import struct
from typing import ClassVar, Self

from .common import OnkyoError
from .instruction import DiscoveryQuery, Instruction
from .message import Message
from .message_code import Code
from .parameter import DestinationArea
from .status import DiscoveryStatus, NotAvailableStatus, RawStatus, Status, status_classes

_LOGGER = logging.getLogger(__name__)


class OnkyoConnectionError(OnkyoError):
    """Onkyo connection error."""


class OnkyoParsingError(OnkyoError):
    """Onkyo parsing error."""


class OnkyoStreamRecoverableError(OnkyoError):
    """Onkyo stream recoverable error."""


class ParseMode(Enum):
    """Parse mode."""

    ISCP = auto()
    EISCP = auto()
    DISCOVERY = auto()


@dataclass
class EISCPPacket[M: Message]:
    """eISCP packet."""

    header: EISCPHeader
    iscp: ISCPData[M]

    @staticmethod
    def parse_status(
        packet_plus: bytes, *, parse_mode: ParseMode = ParseMode.EISCP
    ) -> EISCPPacket[Status]:
        """Parse the eISCP status packet from reader."""
        header = EISCPHeader.parse(packet_plus)
        if len(packet_plus) < header.total_size:
            raise OnkyoParsingError(
                f"Too short eISCP packet: {len(packet_plus)} < {header.total_size}.\n"
                f"Packet plus: {packet_plus!r}"
            )
        data = packet_plus[header.header_size : header.total_size]
        iscp = ISCPStatusData.parse(data, parse_mode=parse_mode)
        return EISCPPacket(header, iscp)

    @staticmethod
    async def parse_status_read(
        reader: asyncio.StreamReader, *, parse_mode: ParseMode = ParseMode.EISCP
    ) -> EISCPPacket[Status]:
        """Parse the eISCP status packet from reader."""
        header = await EISCPHeader.parse_read(reader)
        data = await reader.readexactly(header.data_size)
        iscp = ISCPStatusData.parse(data, parse_mode=parse_mode)
        return EISCPPacket(header, iscp)

    @staticmethod
    def from_instruction_message(
        message: Instruction, unit_type: UnitType | None = None
    ) -> EISCPPacket[Instruction]:
        """Create eISCP packet from instruction message."""
        if unit_type is None:
            data = ISCPInstructionData(message)
        else:
            data = ISCPInstructionData(message, unit_type)
        header = EISCPHeader(data_size=len(data.raw))
        return EISCPPacket(header, data)

    @cached_property
    def raw(self) -> bytes:
        """Serialize the eISCP packet."""
        return self.header.raw + self.iscp.raw

    @property
    def size(self) -> int:
        """Total size of the packet."""
        return self.header.total_size


class OnkyoHeaderError(OnkyoParsingError):
    """Onkyo header parsing error."""


@dataclass(kw_only=True)
class EISCPHeader:
    """eISCP header."""

    basic_length: ClassVar[int] = 16

    header_size: int = basic_length
    data_size: int
    version: int = 1
    reserved: bytes = b"\x00\x00\x00"
    additional: bytes | None = None

    _struct: ClassVar[struct.Struct] = struct.Struct("! 4s I I B 3s")

    @classmethod
    def parse(cls, header_plus: bytes) -> Self:
        """Parse the eISCP header."""
        if len(header_plus) < cls.basic_length:
            raise OnkyoHeaderError(
                f"Way too short eISCP header: {len(header_plus)} < {cls.basic_length}.\n"
                f"Header plus: {header_plus!r}"
            )
        header_basic = header_plus[: cls.basic_length]
        self = cls._parse_basic(header_basic)
        if self.header_size > cls.basic_length:
            if len(header_plus) < self.header_size:
                raise OnkyoHeaderError(
                    f"Too short eISCP header: {len(header_plus)} < {self.header_size}.\n"
                    f"Header plus: {header_plus!r}"
                )
            self.additional = header_plus[cls.basic_length : self.header_size]
        return self

    @classmethod
    async def parse_read(cls, reader: asyncio.StreamReader) -> Self:
        """Parse the eISCP header from reader."""
        basic = await reader.readexactly(cls.basic_length)
        self = cls._parse_basic(basic)
        additional_size = self.header_size - cls.basic_length
        if additional_size:
            self.additional = await reader.readexactly(additional_size)
        return self

    @classmethod
    def _parse_basic(cls, header: bytes) -> Self:
        """Parse the eISCP basic header."""
        try:
            unpacked = cls._struct.unpack(header)
        except struct.error as exc:
            raise OnkyoHeaderError(
                f"Invalid eISCP header structure: {exc!r}.\n"  # str-sep
                f"Header basic: {header!r}"
            ) from None
        magic, header_size, data_size, version, reserved = unpacked

        if magic != b"ISCP":
            raise OnkyoHeaderError(
                f"Invalid eISCP header magic: {magic}.\n"  # str-sep
                f"Header basic: {header!r}"
            )

        return cls(
            header_size=header_size,
            data_size=data_size,
            version=version,
            reserved=reserved,
        )

    @cached_property
    def raw(self) -> bytes:
        """Serialize the eISCP header."""
        packed_basic = self._struct.pack(
            b"ISCP", self.header_size, self.data_size, self.version, self.reserved
        )
        if self.additional is None:
            return packed_basic
        return packed_basic + self.additional

    @cached_property
    def total_size(self) -> int:
        """Get the total size of the header and data."""
        return self.header_size + self.data_size


class UnitType(Enum):
    """Unit type."""

    RECEIVER = b"1"
    DISCOVERY = b"x"
    PIONEER_DISCOVERY = b"p"


@dataclass
class ISCPData[M: Message]:
    """ISCP data."""

    message: M
    unit_type: UnitType
    end: bytes

    @cached_property
    def raw(self) -> bytes:
        """Serialize the ISCP data."""
        # Start character is "!"
        return b"!" + self.unit_type.value + self.message.raw + self.end


class OnkyoISCPDataError(OnkyoParsingError, OnkyoStreamRecoverableError):
    """Onkyo ISCP data parsing error."""


class OnkyoISCPMessageError(OnkyoISCPDataError):
    """Onkyo ISCP message parsing error."""

    raw_status: RawStatus

    def __init__(self, reason: str, raw_status: RawStatus) -> None:
        """Initialize."""
        super().__init__(f"{reason} Raw: {raw_status}")
        self.raw_status = raw_status


@dataclass
class ISCPStatusData(ISCPData[Status]):
    """ISCP status data."""

    @classmethod
    def parse(cls, data: bytes, *, parse_mode: ParseMode = ParseMode.EISCP) -> Self:
        """Parse the ISCP status data."""
        if data[0:1] != b"!":
            raise OnkyoISCPDataError(
                f"Invalid ISCP start: {data[0:1]!r}.\n"  # str-sep
                f"Data: {data!r}"
            )

        end = cls._parse_end(data, parse_mode)
        message_end = -len(end)

        try:
            unit_type = UnitType(data[1:2])
        except ValueError:
            raise OnkyoISCPDataError(
                f"Unknown ISCP unit type: {data[1:2]!r}.\n"  # str-sep
                f"Data: {data!r}"
            ) from None

        message_start = 2
        split = message_start + 3  # 3 is the code size
        code, parameter = data[message_start:split], data[split:message_end]
        message = cls._parse_message(code, parameter)

        return cls(message, unit_type, end)

    @staticmethod
    def _parse_end(data: bytes, parse_mode: ParseMode) -> bytes:
        """Parse the ISCP end characters."""
        ends: tuple[bytes, ...]
        match parse_mode:
            case ParseMode.ISCP:
                ends = (b"\x1a",)  # EOF (SUB)
            case ParseMode.EISCP:
                ends = (
                    b"\x1a\r\n",  # EOF (SUB) followed by CR+LF
                    b"\r\n",  # CR+LF ; example: b'!1TUN\r\n'
                )
            case ParseMode.DISCOVERY:
                ends = (b"\x19\r\n",)  # EOM followed by CR+LF

        for end in ends:
            if data.endswith(end):
                break
        else:
            raise OnkyoISCPDataError(
                f"Invalid ISCP end: ...{data[-5:]!r}. Parse mode: {parse_mode}.\n"  # str-sep
                f"Data: {data!r}"
            )

        return end

    @staticmethod
    def _parse_message(raw_code: bytes, parameter: bytes) -> Status:
        """Parse the ISCP status message."""
        code = Code.parse(raw_code)
        if code is None:
            return RawStatus(raw_code, parameter)
        status_cls = status_classes.get(code.kind)
        if status_cls is None:
            raw_status = RawStatus(raw_code, parameter)
            raise OnkyoISCPMessageError("ISCP message with no matching class.", raw_status)
        if parameter == b"N/A":
            return NotAvailableStatus.parse(code, parameter)
        try:
            status = status_cls.parse(code, parameter)
        except ValueError as exc:
            raw_status = RawStatus(raw_code, parameter)
            raise OnkyoISCPMessageError(
                f"ISCP message that couldn't be parsed ({exc}).", raw_status
            ) from None
        except Exception as exc:
            raw_status = RawStatus(raw_code, parameter)
            raise OnkyoISCPMessageError(
                f"¡ISCP message that couldn't be parsed a lot! ({exc!r})", raw_status
            ) from exc
        return status


@dataclass
class ISCPInstructionData(ISCPData[Instruction]):
    """ISCP instruction data."""

    unit_type: UnitType = UnitType.RECEIVER
    end: bytes = b"\r\n"  # End characters may be CR, LF or CR+LF


def _validate_status_packet(packet: EISCPPacket[Status]) -> Status | None:
    """Validate eISCP status packet."""
    header = packet.header
    iscp = packet.iscp
    # Not erroring here, assuming backwards compatibility in the future
    if header.version != 1:
        _LOGGER.info(
            "Unexpected eISCP header version: %s.\n"  # str-sep
            "Packet: %s",
            header.version,
            packet,
        )
    if header.header_size != EISCPHeader.basic_length:
        _LOGGER.info(
            "Unexpected eISCP header size : %s.\n"  # str-sep
            "Packet: %s",
            header.header_size,
            packet,
        )
    if header.reserved != b"\x00\x00\x00":
        _LOGGER.info(
            "Unexpected eISCP header reserved: %r.\n"  # str-sep
            "Packet: %s",
            header.reserved,
            packet,
        )
    if iscp.unit_type != UnitType.RECEIVER:
        _LOGGER.info(
            "Unexpected eISCP unit type: %s.\n"  # str-sep
            "Packet: %s",
            iscp.unit_type,
            packet,
        )
        return None
    return packet.iscp.message


async def read_message(reader: asyncio.StreamReader) -> Status:
    """Read message."""
    while True:
        try:
            packet = await EISCPPacket.parse_status_read(reader)
        except (OSError, EOFError) as exc:
            raise OnkyoConnectionError(f"{exc!r}") from None
        except OnkyoStreamRecoverableError as exc:
            _LOGGER.error("Invalid response: %s(%s)", type(exc).__name__, exc)
            continue
        except OnkyoParsingError as exc:
            _LOGGER.error("Invalid response (unrecoverable): %s(%s)", type(exc).__name__, exc)
            raise
        except Exception:
            _LOGGER.exception("¡Very invalid response!")
            raise
        message = _validate_status_packet(packet)
        if message is None:
            continue
        return message


async def write_message(writer: asyncio.StreamWriter, message: Instruction) -> None:
    """Write message."""
    packet = EISCPPacket.from_instruction_message(message)
    try:
        writer.write(packet.raw)
        await writer.drain()
    except OSError as exc:
        raise OnkyoConnectionError(f"{exc!r}") from None


@dataclass(kw_only=True)
class DiscoveryInfo:
    """Discovery information."""

    ip: str
    port: int
    model_name: str
    iscp_port: int
    destination_area: DestinationArea
    identifier: str


@dataclass
class EISCPDiscovery(asyncio.DatagramProtocol):
    """eISCP Discovery Protocol."""

    target_str: str
    callback: Callable[[DiscoveryInfo], None]
    transport: asyncio.DatagramTransport | None = field(init=False, default=None)

    _discovered: set[str] = field(default_factory=set)

    packets: ClassVar[list[EISCPPacket[Instruction]]] = [
        EISCPPacket.from_instruction_message(DiscoveryQuery(), UnitType.DISCOVERY),
        EISCPPacket.from_instruction_message(DiscoveryQuery(), UnitType.PIONEER_DISCOVERY),
    ]

    def connection_made(
        self,
        transport: asyncio.DatagramTransport,  # type: ignore[override]
    ) -> None:
        """Send discovery packets after connection made."""
        self.transport = transport

        _LOGGER.debug("Sending discovery packets to %s", self.target_str)
        for packet in self.packets:
            self.transport.sendto(packet.raw)

    @staticmethod
    def _get_message(data: bytes) -> DiscoveryStatus | None:
        try:
            packet = EISCPPacket.parse_status(data, parse_mode=ParseMode.DISCOVERY)
        except OnkyoParsingError:
            _LOGGER.exception(
                "Invalid discovery response.\n"  # str-sep
                "Data: %s",
                data,
            )
            return None
        except Exception:
            _LOGGER.exception(
                "¡Very invalid discovery response!\n"  # str-sep
                "Data: %s",
                data,
            )
            return None
        message = _validate_status_packet(packet)
        if message is None:
            return None
        if not isinstance(message, DiscoveryStatus):
            _LOGGER.warning(
                "Invalid discovery response message type: %s\nPacket: %s",
                type(message).__name__,
                packet,
            )
            return None
        return message

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        """Parse and forward discovery info."""
        ip, port = addr
        _LOGGER.debug("Received discovery response from %s:%d", ip, port)

        message = self._get_message(data)
        if message is None:
            return

        if message.identifier in self._discovered:
            return
        self._discovered.add(message.identifier)

        discovery_info = DiscoveryInfo(
            ip=ip,
            port=port,
            model_name=message.model_name,
            iscp_port=message.iscp_port,
            destination_area=message.destination_area,
            identifier=message.identifier,
        )
        _LOGGER.debug("Discovery info: %s", discovery_info)

        self.callback(discovery_info)

    def close(self) -> None:
        """Close the discovery connection."""
        _LOGGER.debug("Closing the discovery connection %s", self.target_str)
        if self.transport is not None:
            self.transport.close()



"""Status messages (export)."""

from ..status import (
    AudioInformationStatus as AudioInformation,
    ChannelMutingStatus as ChannelMuting,
    DiscoveryStatus as Discovery,
    FLDisplayStatus as FLDisplay,
    HDMIOutputStatus as HDMIOutput,
    InputSourceStatus as InputSource,
    KnownStatus as Known,
    ListeningModeStatus as ListeningMode,
    MutingStatus as Muting,
    NotAvailableStatus as NotAvailable,
    PowerStatus as Power,
    RawStatus as Raw,
    Status as Status,
    ToneStatus as Tone,
    TunerPresetStatus as TunerPreset,
    ValidStatus as Valid,
    VideoInformationStatus as VideoInformation,
    VolumeStatus as Volume,
)



"""Parameters."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import Any, ClassVar, Final, Self, cast, overload

from .message_code import ALL_ZONES, Zone


class _ParamBase:
    """Base class for params, like ParamEnum or ParamNumeric."""

    @property
    def value(self) -> str:
        """Param value."""
        # only for typing, this function is actually removed
        return ""

    @cached_property
    def raw(self) -> bytes:
        """Param value in bytes form."""
        return self.value.encode()


del _ParamBase.value


class _ParamEnumBase(_ParamBase):
    meanings: tuple[str, ...] = ()
    all_meanings: tuple[str, ...] = ()
    zones: tuple[Zone, ...] = ()


@dataclass(slots=True)
class ParamEnumSettings:
    meanings: str | tuple[str, ...] = ()
    extra_meanings: str | tuple[str, ...] = field(kw_only=True, default=())
    zones: tuple[Zone, ...] | None = field(kw_only=True, default=None)


def _get_empty_zone_mapping() -> dict[Zone, set[Any]]:
    return {zone: set() for zone in ALL_ZONES}


_DEFAULT_ZONE_MAPPING = _get_empty_zone_mapping()

_DEFAULT_PARAM_ENUM_SETTINGS: Final = ParamEnumSettings()


class ParamEnum(_ParamEnumBase, Enum):
    # mypy - "Final name declared in class body cannot depend on type variables"
    __zone_mapping: ClassVar[dict[Zone, set[Self]]] = _DEFAULT_ZONE_MAPPING  # type: ignore[misc]

    def __new__(
        cls, value: str, _settings: ParamEnumSettings = _DEFAULT_PARAM_ENUM_SETTINGS, /
    ) -> Self:
        # Only use first argument as the Enum value
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    # Empty overload init below to provide proper signature to mypy.
    # It is necessary, becauese mypy does not understand Enum __new__ magic.
    @overload  # type: ignore[misc] # mypy - "Single overload definition, multiple required"
    def __init__(self, value: str, /) -> None: ...

    def __init__(
        self, value: str, settings: ParamEnumSettings = _DEFAULT_PARAM_ENUM_SETTINGS, /
    ) -> None:
        meanings = settings.meanings
        extra_meanings = settings.extra_meanings
        zones = settings.zones
        if zones is None:
            zones = self._default_settings_zones()
        if isinstance(meanings, str):
            meanings = (meanings,)
        if isinstance(extra_meanings, str):
            extra_meanings = (extra_meanings,)

        # _value2member_map_ is undocumented
        canonical = cast(ParamEnum | None, self._value2member_map_.get(value))  # type: ignore[attr-defined]
        if canonical is not None:
            canonical.meanings = (*canonical.meanings, *meanings)
            canonical.all_meanings = (*canonical.all_meanings, *meanings, *extra_meanings)
            return

        # canonical does not yet exist, so it means we are canonical

        if not meanings:
            meanings = (self.name.replace("_", " "),)
        self.meanings = meanings
        self.all_meanings = (*meanings, *extra_meanings)

        if not zones:
            return

        self.zones = zones
        if self.__zone_mapping is _DEFAULT_ZONE_MAPPING:
            # mypy - 'Cannot assign to final attribute "__zone_mapping"'
            type(self).__zone_mapping = _get_empty_zone_mapping()  # type: ignore[misc] # noqa: SLF001
        for zone in self.zones:
            # mypy - 'Argument 1 to "add" of "set" has incompatible type "ParamEnum"; expected "Self"'
            self.__zone_mapping[zone].add(self)  # type: ignore[arg-type]

    @classmethod
    def _default_settings_zones(cls) -> tuple[Zone, ...]:
        return ALL_ZONES

    @classmethod
    def for_zone(cls, zone: Zone) -> set[Self]:
        return cls.__zone_mapping[zone]

    @classmethod
    def parse(cls, parameter: bytes) -> Self:
        return cls(parameter.decode())


class MainZoneParamEnum(ParamEnum):
    @classmethod
    def _default_settings_zones(cls) -> tuple[Zone, ...]:
        return (Zone.MAIN,)


class PowerParam(ParamEnum):
    STANDBY = "00"
    ON = "01"
    # ALL_STANDBY = "ALL", ParamEnumSettings(zones=(Zone.MAIN,))


class MutingParam(ParamEnum):
    OFF = "00"
    ON = "01"


class VolumeParamEnum(ParamEnum):
    UP = "UP"
    DOWN = "DOWN"
    UP1 = "UP1", ParamEnumSettings("UP 1")
    DOWN1 = "DOWN1", ParamEnumSettings("DOWN 1")


class InputSourceParam(ParamEnum):
    VIDEO1_DVR_VCR_STB = "00", ParamEnumSettings(("VIDEO1", "VCR/DVR", "STB/DVR"))
    VIDEO1 = DVR = VCR = STB = VIDEO1_DVR_VCR_STB[0]
    VIDEO2_CBL_SAT = "01", ParamEnumSettings(("VIDEO2", "CBL/SAT"))
    VIDEO2 = CBL = SAT = VIDEO2_CBL_SAT[0]
    VIDEO3_GAME = "02", ParamEnumSettings(("VIDEO3", "GAME/TV", "GAME"))
    VIDEO3 = GAME = VIDEO3_GAME[0]
    VIDEO4_AUX = "03", ParamEnumSettings(("VIDEO4", "AUX"))
    VIDEO4 = AUX = VIDEO4_AUX[0]
    VIDEO5_GAME2_AUX2 = "04", ParamEnumSettings(("VIDEO5", "AUX2", "GAME2"))
    VIDEO5 = GAME2 = AUX2 = VIDEO5_GAME2_AUX2[0]
    VIDEO6_PC = "05", ParamEnumSettings(("VIDEO6", "PC"))
    VIDEO6 = PC = VIDEO6_PC[0]
    VIDEO7 = "06"
    HIDDEN1_EXTRA1 = "07", ParamEnumSettings(("HIDDEN1", "EXTRA1"))
    HIDDEN2_EXTRA2 = "08", ParamEnumSettings(("HIDDEN2", "EXTRA2"))
    HIDDEN3_EXTRA3 = "09", ParamEnumSettings(("HIDDEN3", "EXTRA3"))
    HIDDEN1 = EXTRA1 = HIDDEN1_EXTRA1[0]
    HIDDEN2 = EXTRA2 = HIDDEN2_EXTRA2[0]
    HIDDEN3 = EXTRA3 = HIDDEN3_EXTRA3[0]
    DVD_BD = "10", ParamEnumSettings(("DVD", "BD/DVD"))
    DVD = BD = DVD_BD[0]
    STRM_BOX = "11", ParamEnumSettings(zones=(Zone.MAIN, Zone.ZONE2, Zone.ZONE3))
    TV = "12", ParamEnumSettings(zones=(Zone.MAIN, Zone.ZONE2, Zone.ZONE3))
    TAPE = "20", ParamEnumSettings(("TAPE", "TV/TAPE"))
    TAPE2 = "21"
    PHONO = "22"
    CD = "23", ParamEnumSettings(("CD", "TV/CD"))
    FM = "24"
    AM = "25"
    TUNER = "26"
    MUSIC_SERVER_P4S_DLNA = "27", ParamEnumSettings(("MUSIC SERVER", "P4S", "DLNA"))
    MUSIC_SERVER = P4S = DLNA = MUSIC_SERVER_P4S_DLNA[0]
    INTERNET_RADIO = "28", ParamEnumSettings(extra_meanings="IRADIO FAVORITE")
    USB = "29", ParamEnumSettings(("USB", "USB (FRONT)"))
    USB_FRONT = USB[0]
    USB_REAR = "2A", ParamEnumSettings("USB (REAR)")
    NETWORK = "2B", ParamEnumSettings(extra_meanings="NET")
    # USB_TOGGLE = "2C", ParamEnumSettings("USB (TOGGLE)")
    AIRPLAY = "2D"
    BLUETOOTH = "2E"
    USB_DAC_IN = "2F", ParamEnumSettings(zones=(Zone.MAIN,))
    MULTI_CH = "30"
    XM = "31"
    SIRIUS = "32"
    DAB = "33"
    UNIVERSAL_PORT = "40"
    LINE = "41", ParamEnumSettings(zones=(Zone.MAIN,))
    LINE2 = "42", ParamEnumSettings(zones=(Zone.MAIN,))
    OPTICAL = "44", ParamEnumSettings(zones=(Zone.MAIN,))
    COAXIAL = "45", ParamEnumSettings(zones=(Zone.MAIN,))
    HDMI_5 = "55", ParamEnumSettings(zones=(Zone.MAIN, Zone.ZONE2))
    HDMI_6 = "56", ParamEnumSettings(zones=(Zone.MAIN, Zone.ZONE2))
    HDMI_7 = "57", ParamEnumSettings(zones=(Zone.MAIN, Zone.ZONE2))
    # OFF = "7F", ParamEnumSettings(zones=(Zone.ZONE2))
    MAIN_SOURCE = "80", ParamEnumSettings(zones=(Zone.ZONE2, Zone.ZONE3, Zone.ZONE4))


class ListeningModeParam(ParamEnum):
    P_00 = "00", ParamEnumSettings("STEREO", zones=(Zone.MAIN, Zone.ZONE2))
    P_01 = "01", ParamEnumSettings("DIRECT", zones=(Zone.MAIN, Zone.ZONE2))
    P_02 = "02", ParamEnumSettings("SURROUND", zones=(Zone.MAIN,))
    P_03 = "03", ParamEnumSettings(("FILM", "GAME RPG", "ADVANCED GAME"), zones=(Zone.MAIN,))
    P_04 = "04", ParamEnumSettings("THX", zones=(Zone.MAIN,))
    P_05 = "05", ParamEnumSettings(("ACTION", "GAME ACTION"), zones=(Zone.MAIN,))
    P_06 = "06", ParamEnumSettings(("MUSICAL", "GAME ROCK", "ROCK/POP"), zones=(Zone.MAIN,))
    P_07 = "07", ParamEnumSettings("MONO MOVIE", zones=(Zone.MAIN,))
    P_08 = "08", ParamEnumSettings(("ORCHESTRA", "CLASSICAL"), zones=(Zone.MAIN,))
    P_09 = "09", ParamEnumSettings("UNPLUGGED", zones=(Zone.MAIN,))
    P_0A = "0A", ParamEnumSettings(("STUDIO MIX", "ENTERTAINMENT SHOW"), zones=(Zone.MAIN,))
    P_0B = "0B", ParamEnumSettings(("TV LOGIC", "DRAMA"), zones=(Zone.MAIN,))
    P_0C = "0C", ParamEnumSettings(("ALL CH STEREO", "EXTENDED STEREO"), zones=(Zone.MAIN,))
    P_0D = (
        "0D",
        ParamEnumSettings(("THEATER DIMENSIONAL", "FRONT STAGE SURROUND"), zones=(Zone.MAIN,)),
    )
    P_0E = (
        "0E",
        ParamEnumSettings(("ENHANCED 7/ENHANCE", "GAME SPORTS", "SPORTS"), zones=(Zone.MAIN,)),
    )
    P_0F = "0F", ParamEnumSettings("MONO", zones=(Zone.MAIN, Zone.ZONE2))
    P_11 = "11", ParamEnumSettings(("PURE AUDIO", "PURE DIRECT"), zones=(Zone.MAIN,))
    P_12 = "12", ParamEnumSettings("MULTIPLEX", zones=(Zone.MAIN, Zone.ZONE2))
    P_13 = "13", ParamEnumSettings(("FULL MONO", "MONO MUSIC"), zones=(Zone.MAIN,))
    P_14 = "14", ParamEnumSettings("DOLBY VIRTUAL/SURROUND ENHANCER", zones=(Zone.MAIN,))
    P_15 = "15", ParamEnumSettings("DTS SURROUND SENSATION", zones=(Zone.MAIN,))
    P_16 = "16", ParamEnumSettings("AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_17 = "17", ParamEnumSettings("DTS VIRTUAL:X", zones=(Zone.MAIN,))
    P_1F = "1F", ParamEnumSettings(("WHOLE HOUSE MODE", "MULTI ZONE MUSIC"), zones=(Zone.MAIN,))
    P_23 = "23", ParamEnumSettings("STAGE (JAPAN GENRE CONTROL)", zones=(Zone.MAIN,))
    P_25 = "25", ParamEnumSettings("ACTION (JAPAN GENRE CONTROL)", zones=(Zone.MAIN,))
    P_26 = "26", ParamEnumSettings("MUSIC (JAPAN GENRE CONTROL)", zones=(Zone.MAIN,))
    P_2E = "2E", ParamEnumSettings("SPORTS (JAPAN GENRE CONTROL)", zones=(Zone.MAIN,))
    P_40 = "40", ParamEnumSettings(("STRAIGHT DECODE", "5.1 CH SURROUND"), zones=(Zone.MAIN,))
    P_41 = "41", ParamEnumSettings("DOLBY EX/DTS ES", zones=(Zone.MAIN,))
    P_42 = "42", ParamEnumSettings("THX CINEMA", zones=(Zone.MAIN,))
    P_43 = "43", ParamEnumSettings("THX SURROUND EX", zones=(Zone.MAIN,))
    P_44 = "44", ParamEnumSettings("THX MUSIC", zones=(Zone.MAIN,))
    P_45 = "45", ParamEnumSettings("THX GAMES", zones=(Zone.MAIN,))
    P_50 = "50", ParamEnumSettings("THX U(2)/S(2)/I/S CINEMA", zones=(Zone.MAIN,))
    P_51 = "51", ParamEnumSettings("THX U(2)/S(2)/I/S MUSIC", zones=(Zone.MAIN,))
    P_52 = "52", ParamEnumSettings("THX U(2)/S(2)/I/S GAMES", zones=(Zone.MAIN,))
    P_80 = (
        "80",
        ParamEnumSettings(("DOLBY ATMOS/DOLBY SURROUND", "PLII/PLIIx MOVIE"), zones=(Zone.MAIN,)),
    )
    P_81 = "81", ParamEnumSettings("PLII/PLIIx MUSIC", zones=(Zone.MAIN,))
    P_82 = "82", ParamEnumSettings(("DTS:X/NEURAL:X", "NEO:6/NEO:X CINEMA"), zones=(Zone.MAIN,))
    P_83 = "83", ParamEnumSettings("NEO:6/NEO:X MUSIC", zones=(Zone.MAIN,))
    P_84 = (
        "84",
        ParamEnumSettings(
            ("DOLBY SURROUND THX CINEMA", "PLII/PLIIx THX CINEMA"), zones=(Zone.MAIN,)
        ),
    )
    P_85 = (
        "85",
        ParamEnumSettings(
            ("DTS NEURAL:X THX CINEMA", "NEO:6/NEO:X THX CINEMA"), zones=(Zone.MAIN,)
        ),
    )
    P_86 = "86", ParamEnumSettings("PLII/PLIIx GAME", zones=(Zone.MAIN,))
    P_87 = (
        "87",
        ParamEnumSettings("NEURAL SURR", extra_meanings="DVS (PI2)", zones=(Zone.MAIN, Zone.ZONE2)),
    )
    P_88 = (
        "88",
        ParamEnumSettings(
            "NEURAL THX/NEURAL SURROUND", extra_meanings="DVS (NEO6)", zones=(Zone.MAIN, Zone.ZONE2)
        ),
    )
    P_89 = (
        "89",
        ParamEnumSettings(("DOLBY SURROUND THX GAMES", "PLII/PLIIx THX GAMES"), zones=(Zone.MAIN,)),
    )
    P_8A = (
        "8A",
        ParamEnumSettings(("DTS NEURAL:X THX GAMES", "NEO:6/NEO:X THX GAMES"), zones=(Zone.MAIN,)),
    )
    P_8B = (
        "8B",
        ParamEnumSettings(("DOLBY SURROUND THX MUSIC", "PLII/PLIIx THX MUSIC"), zones=(Zone.MAIN,)),
    )
    P_8C = (
        "8C",
        ParamEnumSettings(("DTS NEURAL:X THX MUSIC", "NEO:6/NEO:X THX MUSIC"), zones=(Zone.MAIN,)),
    )
    P_8D = "8D", ParamEnumSettings("NEURAL THX CINEMA", zones=(Zone.MAIN,))
    P_8E = "8E", ParamEnumSettings("NEURAL THX MUSIC", zones=(Zone.MAIN,))
    P_8F = "8F", ParamEnumSettings("NEURAL THX GAMES", zones=(Zone.MAIN,))
    P_90 = "90", ParamEnumSettings("PLIIz HEIGHT", zones=(Zone.MAIN,))
    P_91 = "91", ParamEnumSettings("NEO:6 CINEMA DTS SURROUND SENSATION", zones=(Zone.MAIN,))
    P_92 = "92", ParamEnumSettings("NEO:6 MUSIC DTS SURROUND SENSATION", zones=(Zone.MAIN,))
    P_93 = "93", ParamEnumSettings("NEURAL DIGITAL MUSIC", zones=(Zone.MAIN,))
    P_94 = "94", ParamEnumSettings("PLIIz HEIGHT + THX CINEMA", zones=(Zone.MAIN,))
    P_95 = "95", ParamEnumSettings("PLIIz HEIGHT + THX MUSIC", zones=(Zone.MAIN,))
    P_96 = "96", ParamEnumSettings("PLIIz HEIGHT + THX GAMES", zones=(Zone.MAIN,))
    P_97 = "97", ParamEnumSettings("PLIIz HEIGHT + THX U2/S2 CINEMA", zones=(Zone.MAIN,))
    P_98 = "98", ParamEnumSettings("PLIIz HEIGHT + THX U2/S2 MUSIC", zones=(Zone.MAIN,))
    P_99 = "99", ParamEnumSettings("PLIIz HEIGHT + THX U2/S2 GAMES", zones=(Zone.MAIN,))
    P_9A = "9A", ParamEnumSettings("NEO:X GAME", zones=(Zone.MAIN,))
    P_A0 = "A0", ParamEnumSettings("PLIIx/PLII Movie + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_A1 = "A1", ParamEnumSettings("PLIIx/PLII MUSIC + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_A2 = "A2", ParamEnumSettings("PLIIx/PLII GAME + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_A3 = "A3", ParamEnumSettings("NEO:6 CINEMA + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_A4 = "A4", ParamEnumSettings("NEO:6 MUSIC + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_A5 = "A5", ParamEnumSettings("NEURAL SURROUND + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_A6 = "A6", ParamEnumSettings("NEURAL DIGITAL MUSIC + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_A7 = "A7", ParamEnumSettings("DOLBY EX + AUDYSSEY DSX", zones=(Zone.MAIN,))
    P_FF = "FF", ParamEnumSettings("AUTO SURROUND", zones=(Zone.MAIN,))
    # MOVIE = "MOVIE", ParamEnumSettings("MOVIE MODE WRAP-AROUND", zones=(Zone.MAIN,))
    # MUSIC = "MUSIC", ParamEnumSettings("MUSIC MODE WRAP-AROUND", zones=(Zone.MAIN,))
    # GAME = "GAME", ParamEnumSettings("GAME MODE WRAP-AROUND", zones=(Zone.MAIN,))
    # THX = "THX", ParamEnumSettings("THX MODE WRAP-AROUND", zones=(Zone.MAIN,))
    # AUTO = "AUTO", ParamEnumSettings("AUTO MODE WRAP-AROUND", zones=(Zone.MAIN,))
    # SURR = "SURR", ParamEnumSettings("SURR MODE WRAP-AROUND", zones=(Zone.MAIN,))
    # STEREO = "STEREO", ParamEnumSettings("STEREO MODE WRAP-AROUND", zones=(Zone.MAIN,))


class HDMIOutputParam(MainZoneParamEnum):
    NO_ANALOG = "00", ParamEnumSettings(("ANALOG", "NO"))
    NO = ANALOG = NO_ANALOG[0]
    MAIN_YES_HDMI = "01", ParamEnumSettings(("MAIN", "YES", "HDMI"))
    MAIN = YES = HDMI = MAIN_YES_HDMI[0]  # HDMI as opposed to ANALOG
    SUB_HDBASET = "02", ParamEnumSettings(("SUB", "HDBaseT"))
    SUB = HDBASET = SUB_HDBASET[0]
    BOTH = "03", ParamEnumSettings(extra_meanings="MAIN + SUB")
    BOTH_MAIN = "04", ParamEnumSettings("BOTH (MAIN)")
    BOTH_SUB = "05", ParamEnumSettings("BOTH (SUB)")


@dataclass
class ParamNumeric(_ParamBase):
    numeric_range: ClassVar[tuple[int, int]]

    numeric: int
    raw: bytes

    @classmethod
    def from_numeric(cls, numeric: int) -> Self:
        if not cls.numeric_range[0] <= numeric <= cls.numeric_range[1]:
            raise ValueError(f"Param outside of range: {numeric} in {cls.__name__}")
        sign = "+" if cls.numeric_range[0] < 0 and numeric != 0 else ""
        return cls(numeric, f"{numeric:{sign}02X}".encode())

    @classmethod
    def parse(cls, parameter: bytes) -> int:
        numeric = int(parameter, 16)
        if not cls.numeric_range[0] <= numeric <= cls.numeric_range[1]:
            raise ValueError(f"Param outside of range: {numeric} in {cls.__name__}")
        return numeric


class VolumeParamNumeric(ParamNumeric):
    numeric_range = (0, 200)


class ToneParam(ParamNumeric):
    numeric_range = (-10, 10)


class TunerPresetParam(ParamNumeric):
    numeric_range = (0, 40)


class DestinationArea(Enum):
    NORTH_AMERICA = "DX"
    EURASIA = "XX"
    JAPAN = "JJ"


__all__ = [
    "InputSourceParam",
    "ListeningModeParam",
    "HDMIOutputParam",
]



"""Receiver."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from enum import Enum
import logging
from typing import TYPE_CHECKING, Final, Generic, Self, TypeVar

from .common import BasicReceiverInfo, ReceiverInfo
from .instruction import Instruction
from .protocol import (
    DiscoveryInfo,
    EISCPDiscovery,
    OnkyoConnectionError,
    read_message,
    write_message,
)
from .status import Status

_LOGGER = logging.getLogger(__name__)


BROADCAST_ADDRESS: Final = "255.255.255.255"


async def interview(
    host: str,
    *,
    port: int = 60128,
) -> ReceiverInfo:
    """Interview Onkyo Receiver."""

    target_str = f"{host}:{port}"
    _LOGGER.debug("Interviewing receiver: %s", target_str)

    receiver_info_future: asyncio.Future[ReceiverInfo] = asyncio.Future()

    def callback(discovery: DiscoveryInfo) -> None:
        """Receiver interviewed, connection not yet active."""
        if receiver_info_future.done():
            return
        receiver_info = ReceiverInfo(
            host=host,
            ip=discovery.ip,
            port=discovery.iscp_port,
            model_name=discovery.model_name,
            identifier=discovery.identifier,
        )
        receiver_info_future.set_result(receiver_info)

    protocol = EISCPDiscovery(target_str, callback)

    loop = asyncio.get_running_loop()
    await loop.create_datagram_endpoint(lambda: protocol, remote_addr=(host, port))

    try:
        receiver_info = await receiver_info_future
    finally:
        protocol.close()

    _LOGGER.debug("Interviewed receiver %s: %s", target_str, receiver_info)

    return receiver_info


async def discover(
    address: str = BROADCAST_ADDRESS,
    *,
    port: int = 60128,
) -> AsyncGenerator[ReceiverInfo]:
    """Discover Onkyo Receivers."""

    target_str = f"{address}:{port}"
    _LOGGER.debug("Discovering receivers on %s", target_str)

    receivers_discovered: set[str] = set()
    receiver_info_queue: asyncio.Queue[ReceiverInfo] = asyncio.Queue()

    def callback(discovery: DiscoveryInfo) -> None:
        """Receiver discovered, connection not yet active."""
        info = ReceiverInfo(
            host=discovery.ip,
            ip=discovery.ip,
            port=discovery.iscp_port,
            model_name=discovery.model_name,
            identifier=discovery.identifier,
        )
        if info.identifier not in receivers_discovered:
            receivers_discovered.add(info.identifier)
            receiver_info_queue.put_nowait(info)

    protocol = EISCPDiscovery(target_str, callback)

    loop = asyncio.get_running_loop()
    await loop.create_datagram_endpoint(
        lambda: protocol,
        remote_addr=(address, port),
        allow_broadcast=True,
    )

    try:
        while True:
            receiver_info = await receiver_info_queue.get()
            _LOGGER.debug("Discovered receiver on %s: %s", target_str, receiver_info)
            yield receiver_info
    finally:
        protocol.close()


InfoT = TypeVar("InfoT", bound=BasicReceiverInfo, default=ReceiverInfo)


async def _connect_receiver_retry(info: InfoT) -> Receiver[InfoT]:
    """Connect to the receiver, retrying on failure."""
    sleep_time = 10
    sleep_time_max = 180
    while True:
        try:
            return await Receiver.open_connection(info)
        except OSError:
            await asyncio.sleep(sleep_time)
            sleep_time = min(sleep_time * 2, sleep_time_max)


@asynccontextmanager
async def connect(
    info: InfoT, *, retry: bool = False, run_queue: bool = False
) -> AsyncGenerator[Receiver[InfoT]]:
    """Connect to the receiver."""
    _LOGGER.debug(
        "Async context manager connect (retry: %s, run_queue: %s): %s", retry, run_queue, info
    )

    if retry:
        receiver = await _connect_receiver_retry(info)
    else:
        receiver = await Receiver.open_connection(info)

    try:
        if run_queue:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(receiver.run_queue())
                yield receiver
        else:
            yield receiver
    finally:
        receiver.close()


class _ReceiverState(Enum):
    """Receiver state."""

    CONNECTED = "CONNECTED"
    RUNNING_QUEUE = "RUNNING_QUEUE"
    CLOSED = "CLOSED"


@dataclass
class Receiver(Generic[InfoT]):
    """Receiver (connected)."""

    info: InfoT
    _reader: asyncio.StreamReader
    _writer: asyncio.StreamWriter
    _read_lock: asyncio.Lock = field(default_factory=asyncio.Lock)
    _write_lock: asyncio.Lock = field(default_factory=asyncio.Lock)
    _read_queue: asyncio.Queue[Status] | None = None
    _write_queue: asyncio.Queue[Instruction] | None = None
    _state: _ReceiverState = _ReceiverState.CONNECTED

    @classmethod
    async def open_connection(cls, info: InfoT) -> Self:
        """Open connection to the receiver."""
        _LOGGER.debug("Connecting: %s", info)
        reader, writer = await asyncio.open_connection(info.host, info.port)
        _LOGGER.debug("Connected: %s", info)
        return cls(info, reader, writer)

    async def run_queue(self) -> None:
        """Run queue reader/writer."""
        if self._state is not _ReceiverState.CONNECTED:
            raise RuntimeError(
                "Run queue called on receiver not in CONNECTED state, "
                f"current state: {self._state}, info: {self.info}"
            )

        self._read_queue = asyncio.Queue()
        self._write_queue = asyncio.Queue()
        _LOGGER.debug("Run queue starting: %s", self.info)
        self._state = _ReceiverState.RUNNING_QUEUE

        try:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self._read_messages())
                tg.create_task(self._write_messages())
        except* OnkyoConnectionError as exc:
            _LOGGER.warning("Disconnect detected (%s): %s", exc.exceptions, self.info)
        finally:
            _LOGGER.debug("Run queue ending: %s", self.info)
            self._do_close()

    async def _read_messages(self) -> None:
        """Read messages."""
        queue = self._read_queue
        if TYPE_CHECKING:
            assert queue is not None
        async with self._read_lock:
            while True:
                message = await read_message(self._reader)
                _LOGGER.debug("[%s]   << %s", self.info.host, message)
                await queue.put(message)

    async def _write_messages(self) -> None:
        """Write messages."""
        queue = self._write_queue
        if TYPE_CHECKING:
            assert queue is not None
        async with self._write_lock:
            while True:
                message = await queue.get()
                await write_message(self._writer, message)
                _LOGGER.debug("[%s] >>>> %s", self.info.host, message)

    async def read(self) -> Status | None:
        """Read from the receiver."""
        if self._state is _ReceiverState.CLOSED:
            return None

        if self._read_queue is None:
            async with self._read_lock:
                try:
                    message = await read_message(self._reader)
                except OnkyoConnectionError as exc:
                    self._disconnect_from_read_write(exc)
                    return None
                except Exception as exc:
                    self._disconnect_from_read_write(exc)
                    raise
        else:
            try:
                message = await self._read_queue.get()
            except asyncio.QueueShutDown:
                return None

        _LOGGER.debug("[%s] <<<< %s", self.info.host, message)
        return message

    async def write(self, message: Instruction) -> None:
        """Write to the receiver."""
        if self._state is _ReceiverState.CLOSED:
            raise RuntimeError(f"Write called on receiver in CLOSED state, info: {self.info}")

        if self._write_queue is None:
            async with self._write_lock:
                try:
                    await write_message(self._writer, message)
                except Exception as exc:
                    self._disconnect_from_read_write(exc)
                    raise
                _LOGGER.debug("[%s] >>>> %s", self.info.host, message)
        else:
            _LOGGER.debug("[%s] >>   %s", self.info.host, message)
            self._write_queue.put_nowait(message)

    def _disconnect_from_read_write(self, exc: Exception) -> None:
        """Disconnect the receiver (called from read/write)."""
        if self._state is _ReceiverState.CLOSED:
            _LOGGER.debug("Disconnect - already closed: %s", self.info)
            return

        if isinstance(exc, OnkyoConnectionError):
            _LOGGER.warning("Disconnect detected (%s): %s", exc, self.info)
        self._do_close()

    def _do_close(self) -> None:
        """Close connection."""
        if self._state is _ReceiverState.CLOSED:
            raise RuntimeError(f"Do close called on receiver in CLOSED state, info: {self.info}")

        _LOGGER.debug("Closing: %s", self.info)
        self._state = _ReceiverState.CLOSED
        self._writer.close()  # writer closes the whole stream, including the reader
        if self._read_queue is not None:
            self._read_queue.shutdown(immediate=True)

    def close(self) -> None:
        """Close connection."""
        if self._state is _ReceiverState.CLOSED:
            return
        if self._state is _ReceiverState.RUNNING_QUEUE:
            raise RuntimeError(
                f"Close called on receiver in RUNNING_QUEUE state, info: {self.info}"
            )
        self._do_close()


class BasicReceiver(Receiver[BasicReceiverInfo]):
    """Basic receiver (connected)."""


__all__ = [
    "interview",
    "discover",
    "connect",
    "Receiver",
]



"""Common."""

from dataclasses import dataclass


class OnkyoError(Exception):
    """Onkyo Error."""


@dataclass(kw_only=True)
class BasicReceiverInfo:
    """Basic receiver information."""

    host: str
    port: int = 60128


@dataclass(kw_only=True)
class ReceiverInfo(BasicReceiverInfo):
    """Receiver information."""

    ip: str
    model_name: str
    identifier: str


__all__ = [
    "BasicReceiverInfo",
    "ReceiverInfo",
]



"""Messages."""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TypedDict, Unpack

from .message_code import Code, Kind, Zone
from .parameter import ParamEnum


class KnownMessage:
    """Abstract known message.

    Each subclass needs to provide `parameter`.
    Each subclass needs to provide `code` or (`kind` and `zone`).
    """

    parameter: bytes

    @cached_property
    def code(self) -> Code:
        return Code.from_kind_zone(self.kind, self.zone)

    @cached_property
    def zone(self) -> Zone:
        return self.code.zone

    @cached_property
    def kind(self) -> Kind:
        return self.code.kind

    @cached_property
    def raw(self) -> bytes:
        return self.code.raw + self.parameter

    def _validate(self, **kwargs: Unpack[ValidateKwargs]) -> None:
        pass


class ValidateKwargs(TypedDict, total=False):
    param: ParamEnum


@dataclass
class RawMessage:
    raw_code: bytes
    parameter: bytes

    @cached_property
    def raw(self) -> bytes:
        return self.raw_code + self.parameter


type Message = KnownMessage | RawMessage


class CheckParamZonesMixin(KnownMessage):
    def _validate(self, **kwargs: Unpack[ValidateKwargs]) -> None:
        param = kwargs.get("param")
        if param is None:
            raise ValueError(f"No param to check zone {self.zone}: {self}")
        if param.zones and self.zone not in param.zones:
            raise ValueError(f"Invalid zone {self.zone} for param {param}: {self}")
        super()._validate(**kwargs)


# __all__ = [
#     "KnownMessage",
#     "RawMessage",
#     "Message",
# ]



"""Instruction messages."""

from dataclasses import dataclass, field
from typing import ClassVar, TypeAlias, Unpack

from .message import CheckParamZonesMixin, KnownMessage, RawMessage, ValidateKwargs
from .message_code import Kind, Zone
from .parameter import (
    HDMIOutputParam,
    InputSourceParam,
    ListeningModeParam,
    MutingParam,
    ParamEnum,
    PowerParam,
    ToneParam,
    TunerPresetParam,
    VolumeParamEnum,
    VolumeParamNumeric,
)


@dataclass
class _Instruction(KnownMessage):
    zone: Zone = field()

    # _validate needs to be called by each subclass
    def _validate(self, **kwargs: Unpack[ValidateKwargs]) -> None:
        try:
            _ = self.code
        except KeyError:
            raise ValueError(f"Invalid zone {self.zone} for {self.__class__.__name__}") from None
        super()._validate(**kwargs)


@dataclass
class _MainZoneInstructionMixin:
    # cannot be a ClassVar because mypy complains
    # mypy - 'error: Missing positional argument "zone" in call'
    zone: Zone = field(init=False, repr=False, default=Zone.MAIN)


@dataclass
class _Query(_Instruction):
    parameter: bytes = field(init=False, repr=False, default=b"QSTN")

    def __post_init__(self) -> None:
        self._validate()


@dataclass
class PowerQuery(_Query):
    kind: ClassVar[Kind] = Kind.POWER


@dataclass
class MutingQuery(_Query):
    kind: ClassVar[Kind] = Kind.MUTING


@dataclass
class ChannelMutingQuery(_MainZoneInstructionMixin, _Query):
    kind: ClassVar[Kind] = Kind.CHANNEL_MUTING


@dataclass
class VolumeQuery(_Query):
    kind: ClassVar[Kind] = Kind.VOLUME


@dataclass
class ToneQuery(_Query):
    kind: ClassVar[Kind] = Kind.TONE


@dataclass
class InputSourceQuery(_Query):
    kind: ClassVar[Kind] = Kind.INPUT_SOURCE


@dataclass
class ListeningModeQuery(_Query):
    kind: ClassVar[Kind] = Kind.LISTENING_MODE


@dataclass
class HDMIOutputQuery(_MainZoneInstructionMixin, _Query):
    kind: ClassVar[Kind] = Kind.HDMI_OUTPUT


@dataclass
class AudioInformationQuery(_MainZoneInstructionMixin, _Query):
    kind: ClassVar[Kind] = Kind.AUDIO_INFORMATION


@dataclass
class VideoInformationQuery(_MainZoneInstructionMixin, _Query):
    kind: ClassVar[Kind] = Kind.VIDEO_INFORMATION


@dataclass
class TunerPresetQuery(_Query):
    kind: ClassVar[Kind] = Kind.TUNER_PRESET


@dataclass
class FLDisplayQuery(_MainZoneInstructionMixin, _Query):
    kind: ClassVar[Kind] = Kind.FL_DISPLAY


@dataclass
class DiscoveryQuery(_MainZoneInstructionMixin, _Query):
    kind: ClassVar[Kind] = Kind.DISCOVERY


@dataclass
class _ParamCommand[ParamT: ParamEnum](_Instruction):
    kind: ClassVar[Kind]

    param: ParamT

    parameter: bytes = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self.parameter = self.param.raw
        self._validate(param=self.param)


class PowerCommand(_ParamCommand[PowerParam]):
    kind: ClassVar[Kind] = Kind.POWER
    Param: TypeAlias = PowerParam


class MutingCommand(_ParamCommand[MutingParam]):
    kind: ClassVar[Kind] = Kind.MUTING
    Param: TypeAlias = MutingParam


class InputSourceCommand(CheckParamZonesMixin, _ParamCommand[InputSourceParam]):
    kind: ClassVar[Kind] = Kind.INPUT_SOURCE
    Param: TypeAlias = InputSourceParam


class ListeningModeCommand(CheckParamZonesMixin, _ParamCommand[ListeningModeParam]):
    kind: ClassVar[Kind] = Kind.LISTENING_MODE
    Param: TypeAlias = ListeningModeParam


# needs to be a dataclass because it inherits from two dataclasses
@dataclass
class _HDMIOutputCommand(_MainZoneInstructionMixin, _ParamCommand[HDMIOutputParam]):
    kind: ClassVar[Kind] = Kind.HDMI_OUTPUT


class HDMIOutputCommand(_HDMIOutputCommand):
    # TypeAlias doesn't work in dataclasses
    Param: TypeAlias = HDMIOutputParam


@dataclass
class _VolumeCommand(_Instruction):
    kind: ClassVar[Kind] = Kind.VOLUME

    param: VolumeParamEnum | int

    parameter: bytes = field(init=False, repr=False)

    def __post_init__(self) -> None:
        if isinstance(self.param, VolumeParamEnum):
            self.parameter = self.param.raw
            self._validate(param=self.param)
            return

        self.parameter = VolumeParamNumeric.from_numeric(self.param).raw
        self._validate()


class VolumeCommand(_VolumeCommand):
    # TypeAlias doesn't work in dataclasses
    Param: TypeAlias = VolumeParamEnum


@dataclass
class TunerPresetCommand(_Instruction):
    kind: ClassVar[Kind] = Kind.TUNER_PRESET

    param: int

    parameter: bytes = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self.parameter = TunerPresetParam.from_numeric(self.param).raw
        self._validate()


@dataclass(kw_only=True)
class ToneCommand(_Instruction):
    kind: ClassVar[Kind] = Kind.TONE

    bass: int | None = None
    treble: int | None = None

    parameter: bytes = field(init=False, repr=False)

    def __post_init__(self) -> None:
        if (self.bass is None and self.treble is None) or (
            self.bass is not None and self.treble is not None
        ):
            raise ValueError(
                f"Exactly one of bass or treble must be set for {self.__class__.__name__}"
            )

        if self.bass is not None:
            self.parameter = b"B" + ToneParam.from_numeric(self.bass).raw

        if self.treble is not None:
            self.parameter = b"T" + ToneParam.from_numeric(self.treble).raw

        self._validate()


# needs to be a dataclass because it inherits from two dataclasses
@dataclass(kw_only=True)
class _ChannelMutingCommand(_MainZoneInstructionMixin, _Instruction):
    kind: ClassVar[Kind] = Kind.CHANNEL_MUTING

    front_left: MutingParam = MutingParam.OFF
    front_right: MutingParam = MutingParam.OFF
    center: MutingParam = MutingParam.OFF
    surround_left: MutingParam = MutingParam.OFF
    surround_right: MutingParam = MutingParam.OFF
    surround_back_left: MutingParam = MutingParam.OFF
    surround_back_right: MutingParam = MutingParam.OFF
    subwoofer: MutingParam = MutingParam.OFF
    height_1_left: MutingParam = MutingParam.OFF
    height_1_right: MutingParam = MutingParam.OFF
    height_2_left: MutingParam = MutingParam.OFF
    height_2_right: MutingParam = MutingParam.OFF
    subwoofer_2: MutingParam = MutingParam.OFF

    parameter: bytes = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self.parameter = b"".join(
            param.raw for param in vars(self).values() if isinstance(param, MutingParam)
        )
        self._validate()


class ChannelMutingCommand(_ChannelMutingCommand):
    # TypeAlias doesn't work in dataclasses
    Param: TypeAlias = MutingParam


type KnownInstruction = (
    PowerQuery
    | MutingQuery
    | ChannelMutingQuery
    | VolumeQuery
    | ToneQuery
    | InputSourceQuery
    | ListeningModeQuery
    | HDMIOutputQuery
    | AudioInformationQuery
    | VideoInformationQuery
    | TunerPresetQuery
    | FLDisplayQuery
    | DiscoveryQuery
    | PowerCommand
    | MutingCommand
    | InputSourceCommand
    | ListeningModeCommand
    | HDMIOutputCommand
    | VolumeCommand
    | TunerPresetCommand
    | ToneCommand
    | ChannelMutingCommand
)


class RawInstruction(RawMessage):
    pass


type Instruction = KnownInstruction | RawInstruction



"""Status messages."""

from collections.abc import Iterable
from dataclasses import dataclass, field
import re
import typing
from typing import ClassVar, Self, TypeAlias

from .message import CheckParamZonesMixin, KnownMessage, RawMessage
from .message_code import Code, Kind, Zone
from .parameter import (
    DestinationArea,
    HDMIOutputParam,
    InputSourceParam,
    ListeningModeParam,
    MutingParam,
    ParamEnum,
    ParamNumeric,
    PowerParam,
    ToneParam,
    TunerPresetParam,
    VolumeParamNumeric,
)


@dataclass
class _KnownStatus(KnownMessage):
    zone: Zone = field(init=False)  # actual implementation in super class, here only for repr
    code: Code = field(repr=False)
    parameter: bytes = field(repr=False)


@dataclass
class _ParamStatus[ParamT: ParamEnum](_KnownStatus):
    kind: ClassVar[Kind]

    Param: ClassVar[type[ParamT]]  # type: ignore[misc] # "ClassVar cannot contain type variables"
    param: ParamT

    __match_args__ = ("param",)

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        param = cls.Param.parse(parameter)
        self = cls(code, parameter, param)
        self._validate(param=param)
        return self


class PowerStatus(_ParamStatus[PowerParam]):
    kind: ClassVar[Kind] = Kind.POWER
    Param: TypeAlias = PowerParam


class MutingStatus(_ParamStatus[MutingParam]):
    kind: ClassVar[Kind] = Kind.MUTING
    Param: TypeAlias = MutingParam


class InputSourceStatus(CheckParamZonesMixin, _ParamStatus[InputSourceParam]):
    kind: ClassVar[Kind] = Kind.INPUT_SOURCE
    Param: TypeAlias = InputSourceParam


class ListeningModeStatus(CheckParamZonesMixin, _ParamStatus[ListeningModeParam]):
    kind: ClassVar[Kind] = Kind.LISTENING_MODE
    Param: TypeAlias = ListeningModeParam


class HDMIOutputStatus(_ParamStatus[HDMIOutputParam]):
    kind: ClassVar[Kind] = Kind.HDMI_OUTPUT
    Param: TypeAlias = HDMIOutputParam


@dataclass
class _NumericStatus(_KnownStatus):
    kind: ClassVar[Kind]

    Param: ClassVar[ParamNumeric]
    param: int

    __match_args__ = ("param",)

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        param = cls.Param.parse(parameter)
        self = cls(code, parameter, param)
        self._validate()
        return self


class VolumeStatus(_NumericStatus):
    kind: ClassVar[Kind] = Kind.VOLUME
    Param: TypeAlias = VolumeParamNumeric


class TunerPresetStatus(_NumericStatus):
    kind: ClassVar[Kind] = Kind.TUNER_PRESET
    Param: TypeAlias = TunerPresetParam


@dataclass
class _ToneStatus(_KnownStatus):
    kind: ClassVar[Kind] = Kind.TONE

    bass: int
    treble: int

    __match_args__ = ("bass", "treble")

    _regex: ClassVar[re.Pattern[bytes]] = re.compile(rb"B(?P<bass>.+?)T(?P<treble>.+)")

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        match = cls._regex.fullmatch(parameter)

        if match is None:
            raise ValueError(f"Regex match fail in {cls.__name__}")

        bass = ToneParam.parse(match["bass"])
        treble = ToneParam.parse(match["treble"])

        self = cls(code, parameter, bass, treble)
        self._validate()
        return self


class ToneStatus(_ToneStatus):
    # TypeAlias doesn't work in dataclasses
    Param: TypeAlias = ToneParam


@dataclass(match_args=False)
class _ChannelMutingStatus(_KnownStatus):
    kind: ClassVar[Kind] = Kind.CHANNEL_MUTING

    front_left: MutingParam
    front_right: MutingParam
    center: MutingParam
    surround_left: MutingParam
    surround_right: MutingParam
    surround_back_left: MutingParam
    surround_back_right: MutingParam
    subwoofer: MutingParam
    height_1_left: MutingParam
    height_1_right: MutingParam
    height_2_left: MutingParam
    height_2_right: MutingParam
    subwoofer_2: MutingParam

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        if len(parameter) != 13 * 2:
            raise ValueError(f"Incorrect number of values in {cls.__name__}")

        values = (MutingParam.parse(parameter[i : i + 2]) for i in range(0, len(parameter), 2))

        self = cls(code, parameter, *values)
        self._validate()
        return self


class ChannelMutingStatus(_ChannelMutingStatus):
    # TypeAlias doesn't work in dataclasses
    Param: TypeAlias = MutingParam


@dataclass(match_args=False)
class AudioInformationStatus(_KnownStatus):
    kind: ClassVar[Kind] = Kind.AUDIO_INFORMATION

    audio_input_port: str | None = None
    input_signal_format: str | None = None
    input_frequency: str | None = None
    input_channels: str | None = None
    listening_mode: str | None = None
    output_channels: str | None = None
    output_frequency: str | None = None
    precision_quartz_lock_system: str | None = None
    auto_phase_control_delay: str | None = None
    auto_phase_control_phase: str | None = None
    upmix_mode: str | None = None

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        values = parameter.split(b",")

        if values[-1] != b"":
            raise ValueError(f"No trailing comma in {cls.__name__}")
        values = values[:-1]

        if len(values) == 0:
            raise ValueError(f"No values in {cls.__name__}")
        if len(values) > 11:
            raise ValueError(f"Too many values in {cls.__name__}")

        decoded_values = [value.decode() for value in values]
        self = cls(code, parameter, *decoded_values)
        self._validate()
        return self


@dataclass(match_args=False)
class VideoInformationStatus(_KnownStatus):
    kind: ClassVar[Kind] = Kind.VIDEO_INFORMATION

    video_input_port: str | None = None
    input_resolution: str | None = None
    input_color_schema: str | None = None
    input_color_depth: str | None = None
    video_output_port: str | None = None
    output_resolution: str | None = None
    output_color_schema: str | None = None
    output_color_depth: str | None = None
    picture_mode: str | None = None
    input_hdr: str | None = None

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        values = parameter.split(b",")

        if values[-1] != b"":
            raise ValueError(f"No trailing comma in {cls.__name__}")
        values = values[:-1]

        if len(values) == 0:
            raise ValueError(f"No values in {cls.__name__}")
        if len(values) > 10:
            raise ValueError(f"Too many values in {cls.__name__}")

        decoded_values = [value.decode() for value in values]
        self = cls(code, parameter, *decoded_values)
        self._validate()
        return self


@dataclass
class FLDisplayStatus(_KnownStatus):
    kind: ClassVar[Kind] = Kind.FL_DISPLAY

    param: str

    __match_args__ = ("param",)

    _special_chars: ClassVar[dict[int, str]] = {
        0x08: "↑",
        0x09: "↓",
        0x0A: "←",
        0x0B: "→",
        0x12: "◗",  # Dolby Logo Left
        0x13: "◖",  # Dolby Logo Right
        # 0x14: "Ⅱ",  # "⏸"?
        # 0x15: "Ω",  # "⏹"?
        0x1A: "⏵",
        0x1B: "⏸",
        0x7F: "▮",  # Cursor
        0x80: "👤",  # Artist
        0x81: "🖸",  # Album
        0x82: "🗀",  # Folder
        0x83: "♫",  # Song
        0x84: "Tr̄",  # Track
        0x90: ".⁰",
        0x91: ".¹",
        0x92: ".²",
        0x93: ".³",
        0x94: ".⁴",
        0x95: ".⁵",
        0x96: ".⁶",
        0x97: ".⁷",
        0x98: ".⁸",
        0x99: ".⁹",
    }

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        if len(parameter) % 2 != 0:
            raise ValueError(f"Odd number of characters {len(parameter)} in {cls.__name__}")

        pairs = (parameter[i : i + 2] for i in range(0, len(parameter), 2))

        special_chars = cls._special_chars

        def translate() -> Iterable[str]:
            for pair in pairs:
                pair_int = int(pair, 16)
                if 0x20 <= pair_int <= 0x7E or 0xA1 <= pair_int <= 0xFF:
                    yield chr(pair_int)
                elif pair_int in special_chars:
                    yield special_chars[pair_int]
                else:
                    raise ValueError(f"Invalid character {pair!r} in {cls.__name__}")

        text = "".join(translate())
        self = cls(code, parameter, text)
        self._validate()
        return self


@dataclass(match_args=False)
class DiscoveryStatus(_KnownStatus):
    kind: ClassVar[Kind] = Kind.DISCOVERY

    model_name: str
    iscp_port: int
    destination_area: DestinationArea
    identifier: str

    _regex: ClassVar[re.Pattern[bytes]] = re.compile(
        rb"""
            (?P<model_name>[^/]{1,64})/
            (?P<iscp_port>\d{5})/
            (?P<destination_area>[A-Z]{2})/
            (?P<identifier>.{1,12})
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        match = cls._regex.fullmatch(parameter)

        if match is None:
            raise ValueError(f"Regex match fail in {cls.__name__}")

        self = cls(
            code=code,
            parameter=parameter,
            model_name=match["model_name"].decode(),
            iscp_port=int(match["iscp_port"]),
            destination_area=DestinationArea(match["destination_area"].decode()),
            identifier=match["identifier"].decode(),
        )
        self._validate()
        return self


@dataclass
class NotAvailableStatus(_KnownStatus):
    kind: Kind = field()

    __match_args__ = ("kind",)

    @classmethod
    def parse(cls, code: Code, parameter: bytes) -> Self:
        self = cls(code, parameter, code.kind)
        self._validate()
        return self


type ValidStatus = (
    PowerStatus
    | MutingStatus
    | InputSourceStatus
    | ListeningModeStatus
    | HDMIOutputStatus
    | VolumeStatus
    | TunerPresetStatus
    | ToneStatus
    | ChannelMutingStatus
    | AudioInformationStatus
    | VideoInformationStatus
    | FLDisplayStatus
    | DiscoveryStatus
)

type KnownStatus = ValidStatus | NotAvailableStatus


class RawStatus(RawMessage):
    pass


type Status = KnownStatus | RawStatus


_status_classes_list: tuple[ValidStatus, ...] = typing.get_args(ValidStatus.__value__)
status_classes = {cls.kind: cls for cls in _status_classes_list}



"""Async Onkyo Python Library."""

# ruff: noqa: F401, F403

from .common import *
from .export import instruction, status
from .export.instruction import Instruction, command, query
from .export.status import Status
from .message_code import *
from .parameter import *
from .receiver import *
